# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import hashlib
import string

import boto3  # pylint: disable=E0401
import botocore  # pylint: disable=E0401
from k8t import config

try:
    from secrets import SystemRandom
except ImportError:
    from random import SystemRandom


LOGGER = logging.getLogger(__name__)
RANDOM_STORE = {}
DEFAULT_SSM_PREFIX = ""
DEFAULT_SSM_REGION = "eu-central-1"


def ssm(key: str, length: int = None) -> str:
    prefix = str(config.CONFIG.get("secrets", {}).get("prefix", DEFAULT_SSM_PREFIX))
    key = prefix + key
    LOGGER.debug("Requesting secret from %s", key)

    region = str(config.CONFIG.get("secrets", {}).get("region", DEFAULT_SSM_REGION))
    client = boto3.client("ssm", region_name=region)

    try:
        result = client.get_parameter(Name=key, WithDecryption=True)["Parameter"][
            "Value"
        ]

        if length is not None:
            if len(result) != length:
                raise AssertionError(
                    "Secret '{}' did not have expected length of {}".format(key, length)
                )

        return result
    except (
        client.exceptions.ParameterNotFound,
        botocore.exceptions.ClientError,
    ) as exc:
        raise RuntimeError(f"Failed to retrieve secret {key}: {exc}") from exc


def random(key: str, length: int = None) -> str:
    LOGGER.debug("Requesting secret from %s", key)

    if key not in RANDOM_STORE:
        RANDOM_STORE[key] = "".join(
            SystemRandom().choice(string.ascii_lowercase + string.digits)

            for _ in range(length or SystemRandom().randint(12, 32))
        )

    if length is not None:
        if len(RANDOM_STORE[key]) != length:
            raise AssertionError(
                "Secret '{}' did not have expected length of {}".format(key, length)
            )

    return RANDOM_STORE[key]


def hash(key: str, length: int = None) -> str:
    LOGGER.debug("Requesting secret from %s", key)

    if key not in RANDOM_STORE:
        hashed_key = hashlib.sha1(key.encode()).hexdigest()
        RANDOM_STORE[key] = hashed_key[:length] if length is not None else hashed_key

    return RANDOM_STORE[key]



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import logging

from typing import List
from jinja2 import Environment, DictLoader, FileSystemLoader, StrictUndefined

from k8t.filters import (b64decode, b64encode, envvar, get_secret, hashf,
                         random_password, sanitize_label, to_bool)
from k8t.project import find_files
from k8t.util import read_file

LOGGER = logging.getLogger(__name__)


def build(path: str, cluster: str, environment: str, template_overrides: List[str] = None) -> Environment:
    env = None
    template_paths = []

    LOGGER.debug(
        "building template environment")

    if template_overrides is not None and len(template_overrides) > 0:
        template_paths = {key: read_file(os.path.abspath(path)) for key, path in template_overrides}

        env = Environment(undefined=StrictUndefined, loader=DictLoader(template_paths))
    else:
        template_paths = find_template_paths(path, cluster, environment)

        env = Environment(undefined=StrictUndefined, loader=FileSystemLoader(template_paths))

    # Filter functions
    env.filters["b64decode"] = b64decode
    env.filters["b64encode"] = b64encode
    env.filters["hash"] = hashf
    env.filters["bool"] = to_bool
    env.filters["sanitize_label"] = sanitize_label

    # Global functions
    env.globals["random_password"] = random_password
    env.globals["get_secret"] = get_secret
    env.globals["env"] = envvar

    # Disabled
    # env.globals['include_raw'] = include_file
    # env.globals['include_file'] = include_file

    return env


def find_template_paths(path: str, cluster: str, environment: str):
    LOGGER.debug(
        "finding template paths in %s for cluster=%s on environment=%s", path, cluster, environment
    )

    template_paths = find_files(
        path, cluster, environment, 'templates', file_ok=False)

    LOGGER.debug(
        "found template paths: %s", template_paths)

    return reversed(template_paths)



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import copy
import json
import logging
import os
import shutil
from functools import reduce
from typing import Any, Dict, List, Tuple

import yaml  # pylint: disable=E0401
from click import secho  # pylint: disable=E0401

from simple_tools.interaction import confirm  # pylint: disable=E0401

LOGGER = logging.getLogger(__name__)


def touch(fname: str, mode=0o666, dir_fd=None, **kwargs) -> None:
    if os.path.exists(fname):
        secho(f"File exists: {fname}", fg="yellow")
    else:
        flags = os.O_CREAT | os.O_APPEND
        with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as filep:
            os.utime(
                filep.fileno() if os.utime in os.supports_fd else fname,
                dir_fd=None if os.supports_fd else dir_fd,
                **kwargs,
            )
        secho(f"File created: {fname}", fg="green")


def makedirs(path, warn_exists=True):
    if os.path.exists(path):
        if warn_exists:
            if confirm("directory {} already exists, go ahead?".format(path)):
                secho(f"Directory exists: {path}", fg="yellow")

                return
            raise RuntimeError("aborting")
        secho(f"Directory exists: {path}", fg="yellow")
    else:
        os.makedirs(path, exist_ok=True)
        secho(f"Directory created: {path}", fg="green")


def replace(source: str, dest: str):
    if os.path.exists(dest):
        if not confirm(f"file {dest} already exists, overwrite?"):
            raise RuntimeError("aborting")

    shutil.copyfile(source, dest)
    secho(f"Template created: {dest}", fg="green")


MERGE_METHODS = ["ltr", "rtl", "ask", "crash"]


def merge(d_1: dict, d_2: dict, path=None, method="ltr"):
    d_1 = copy.deepcopy(d_1)

    if path is None:
        path = []

    for key in d_2:
        if key in d_1:
            if isinstance(d_1[key], dict) and isinstance(d_2[key], dict):
                d_1[key] = merge(d_1[key], d_2[key], path + [str(key)], method=method)
            elif d_1[key] == d_2[key]:
                pass  # same leaf value
            else:
                if method == "ltr":
                    d_1[key] = d_2[key]
                elif method == "rtl":
                    pass
                elif method == "ask":
                    raise NotImplementedError('Merge method "ask"')
                elif method == "crash":
                    raise Exception("Conflict at %s" % ".".join(path + [str(key)]))
                else:
                    raise Exception("Invalid merge method: %s" % method)
        else:
            d_1[key] = d_2[key]

    return d_1


def deep_merge(*dicts, method="ltr"):
    LOGGER.debug('"%s" merging %s dicts', method, len(dicts))

    if not dicts:
        return {}

    return reduce(
        lambda a, b: merge(a, b, method=method) if b is not None else a, dicts
    )


def load_yaml(path: str) -> dict:
    LOGGER.debug("loading values file: %s", path)

    with open(path, "r") as stream:
        return yaml.safe_load(stream) or dict()


def load_cli_value(key: str, value: str) -> Tuple[str, Any]:
    LOGGER.debug("loading cli value (%s, %s)", key, value)

    try:
        return key, json.loads(value, parse_float=lambda x: str(x))  # lambda is necessary, pylint: disable=W0108
    except json.decoder.JSONDecodeError:
        return key, value


def envvalues() -> Dict:
    prefix: str = "K8T_VALUE_"
    values: dict = dict()

    for key, value in os.environ.items():
        if key.startswith(prefix):
            values[key.replace(prefix, "", 1).lower()] = value

    return values


def list_files(
    directory: str, include_files=False, include_directories=False
) -> List[str]:
    result = []

    for _, dirs, files in os.walk(directory):
        if include_files:
            result.extend(files)

        if include_directories:
            result.extend(dirs)

        break

    return result

def read_file(path: str) -> str:
    with open(path, 'rb') as stream:
        return stream.read().decode()



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import os
import sys
from functools import update_wrapper

import click
from jinja2.exceptions import UndefinedError

import coloredlogs
import k8t
from k8t import cluster, config, environment, project, scaffolding, values
from k8t.engine import build
from k8t.templates import YamlValidationError, analyze, render, validate
from k8t.util import (MERGE_METHODS, deep_merge, envvalues, load_cli_value,
                      load_yaml)


def requires_project_directory(func):
    @click.pass_context
    def new_func(ctx, *args, **kwargs):
        if not project.check_directory(kwargs["directory"]):
            sys.exit("not a valid project: {}".format(kwargs["directory"]))

        return ctx.invoke(func, *args, **kwargs)

    return update_wrapper(new_func, func)


@click.group()
@click.version_option(version=k8t.__version__)
@click.option("-d", "--debug", is_flag=True, default=False, show_default=True, help="Enable debug logging.")
@click.option("-t", "--trace", is_flag=True, default=False, show_default=True, help="Enable spammy logging.")
def root(debug, trace):
    coloredlogs.install(level=logging.DEBUG if debug else logging.INFO)

    if not trace:
        logging.getLogger("botocore").setLevel(
            logging.WARN if not debug else logging.INFO)
        logging.getLogger("urllib3").setLevel(
            logging.WARN if not debug else logging.INFO)


@root.command(name="license", help="Print software license.")
def print_license():
    click.echo(k8t.__license__)

# pylint: disable=too-many-locals,too-many-arguments
@root.command(name="validate", help="Validate template files for given context.")
@click.option("-m", "--method", type=click.Choice(MERGE_METHODS), default="ltr", show_default=True, help="Value file merge method.")
@click.option("--value-file", "value_files", multiple=True, type=click.Path(dir_okay=False, exists=True), help="Additional value file to include.")
@click.option("--value", "cli_values", type=(str, str), multiple=True, metavar="KEY VALUE", help="Additional value(s) to include.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.option("--environment", "-e", "ename", metavar="NAME", help="Deployment environment to use.")
@click.option("--suffix", "-s", "suffixes", default=[".yaml", ".j2", ".jinja2"], help="Filter template files by suffix. Can be used multiple times.", show_default=True)
@click.option("--template-file", "-t", "template_overrides", metavar="KEY PATH", type=click.Tuple([str, str]), multiple=True, help="Restrict validation to single template file (the key is needed for references in templates).")
@click.argument("directory", type=click.Path(dir_okay=True, file_okay=False, exists=True), default=os.getcwd())
@requires_project_directory
def cli_validate(method, value_files, cli_values, cname, ename, suffixes, template_overrides, directory):
    vals = deep_merge(  # pylint: disable=redefined-outer-name
        values.load_all(directory, cname, ename, method),
        *(load_yaml(p) for p in value_files),
        dict(load_cli_value(k, v) for k, v in cli_values),
        envvalues(),
        method=method,
    )
    config.CONFIG = config.load_all(directory, cname, ename, method)

    eng = build(directory, cname, ename, template_overrides)

    templates = eng.list_templates()  # pylint: disable=redefined-outer-name

    if suffixes:
        templates = [name for name in templates if os.path.splitext(name)[1] in suffixes]

    all_validated = True

    for template_path in templates:
        errors = set()

        undefined, _, invalid, secrets = analyze(template_path, vals, eng)

        if undefined or invalid:
            for var in undefined:
                errors.add("undefined variable: {}".format(var))

            for var in invalid:
                errors.add("invalid variable: {}".format(var))

        if secrets:
            if "secrets" not in config.CONFIG or "provider" not in config.CONFIG["secrets"]:
                errors.add("No secrets provider configured")

        if errors:
            all_validated = False

            click.secho("{}: ✗".format(template_path), fg="red", err=True)

            for error in errors:
                click.echo("- {}".format(error))
        else:
            click.secho("{}: ✔".format(template_path), fg="green")

    sys.exit(not all_validated)


@root.command(name="gen", help="Create manifest files using stored templates.")
@click.option("-m", "--method", type=click.Choice(MERGE_METHODS), default="ltr", show_default=True, help="Value file merge method.")
@click.option("--value-file", "value_files", multiple=True, type=click.Path(dir_okay=False, exists=True), help="Additional value file to include.")
@click.option("--value", "cli_values", type=(str, str), multiple=True, metavar="KEY VALUE", help="Additional value(s) to include.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.option("--environment", "-e", "ename", metavar="NAME", help="Deployment environment to use.")
@click.option("--suffix", "-s", "suffixes", default=[".yaml", ".j2", ".jinja2"], help="Filter template files by suffix. Can be used multiple times.", show_default=True)
@click.option("--secret-provider", help="Secret provider override.", type=click.Choice(['ssm', 'random', 'hash']))
@click.option("--template-file", "-t", "template_overrides", metavar="KEY PATH", type=click.Tuple([str, str]), multiple=True, help="Restrict validation to single template file (the key is needed for references in templates).")
@click.argument("directory", type=click.Path(dir_okay=True, file_okay=False, exists=True), default=os.getcwd())
@requires_project_directory
def cli_gen(method, value_files, cli_values, cname, ename, suffixes, secret_provider, template_overrides, directory):  # pylint: disable=redefined-outer-name,too-many-arguments
    vals = deep_merge(  # pylint: disable=redefined-outer-name
        values.load_all(directory, cname, ename, method),
        *(load_yaml(p) for p in value_files),
        dict(load_cli_value(k, v) for k, v in cli_values),
        envvalues(),
        method=method,
    )

    config.CONFIG = config.load_all(directory, cname, ename, method)

    if secret_provider is not None:
        if 'secrets' not in config.CONFIG:
            config.CONFIG['secrets'] = dict()

        config.CONFIG['secrets']['provider'] = secret_provider

    eng = build(directory, cname, ename, template_overrides)

    templates = eng.list_templates()  # pylint: disable=redefined-outer-name

    if suffixes:
        templates = [name for name in templates if os.path.splitext(name)[1] in suffixes]

    validated = True

    for template_path in templates:
        if not validate(template_path, vals, eng):
            click.echo("Failed to validate template {}".format(template_path))

            validated = False

    if not validated:
        sys.exit(1)

    try:
        for template_path in templates:
            click.echo("---")
            click.echo("# Source: {}".format(template_path))

            template_output = render(template_path, vals, eng)
            click.echo(template_output)
    except (UndefinedError, YamlValidationError) as err:
        click.secho("✗ -> {}".format(err), fg="red", err=True)
        sys.exit(1)


@root.group(help="Code scaffolding commands.")
def new():
    pass


@new.command(name="project", help="Create a new project.")
@click.argument("directory", type=click.Path())
def new_project(directory):
    scaffolding.new_project(directory)


@new.command(name="cluster", help="Create a new cluster context.")
@click.argument("name")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def new_cluster(name, directory):
    scaffolding.new_cluster(directory, name)


@new.command(name="environment", help="Create a new environment context.")
@click.option("--cluster", "-c", "cname", metavar="NAME")
@click.argument("name")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def new_environment(cname, name, directory):
    base_path = project.get_base_dir(directory, cname, environment=None)

    scaffolding.new_environment(base_path, name)


@new.command(name="template", help="Create specified kubernetes manifest template.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.option("--environment", "-e", "ename", metavar="NAME", help="Deployment environment to use.")
@click.option("--name", "-n", help="Template filename.")
@click.option("--prefix", "-p", help="Prefix for filename.")
@click.argument("kind", type=click.Choice(sorted(list(scaffolding.list_available_templates()))))
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def new_template(cname, ename, name, prefix, kind, directory):
    base_path = project.get_base_dir(directory, cname, ename)

    template_name = "{0}{1}{2}{3}.yaml.j2".format(
        prefix or '',
        name or '',
        '-' if name else '',
        kind
    )

    scaffolding.new_template(base_path, template_name, kind)


@root.group(help="Project inspection commands.")
def get():
    pass


@get.command(name="clusters", help="Get configured clusters.")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def get_clusters(directory):
    for cluster_path in cluster.list_all(directory):
        click.echo(cluster_path)


@get.command(name="environments", help="Get configured environments.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def get_environments(cname, directory):  # pylint: disable=redefined-outer-name
    path = project.get_base_dir(directory, cname, environment=None)

    for environment_path in environment.list_all(path):
        click.echo(environment_path)


@get.command(name="templates", help="Get stored templates.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.option("--environment", "-e", "ename", metavar="NAME", help="Deployment environment to use.")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def get_templates(directory, cname, ename):  # pylint: disable=redefined-outer-name
    for template_path in build(directory, cname, ename).list_templates():
        click.echo(template_path)


@root.group(help="Edit local project files.")
def edit():
    pass


@edit.command(name="config", help="Edit config files in chosen context.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.option("--environment", "-e", "ename", metavar="NAME", help="Deployment environment to use.")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def edit_config(directory, cname, ename):  # pylint: disable=redefined-outer-name
    file_path: str

    if cname is not None:
        if ename is not None:
            file_path = os.path.join(
                directory, "clusters", cname, "environments", ename, "config.yaml"
            )
        else:
            file_path = os.path.join(
                directory, "clusters", cname, "config.yaml")
    else:
        file_path = os.path.join(directory, "config.yaml")

    click.edit(filename=file_path)


@edit.command(name="values", help="Edit value files in chosen context.")
@click.option("--cluster", "-c", "cname", metavar="NAME", help="Cluster context to use.")
@click.option("--environment", "-e", "ename", metavar="NAME", help="Deployment environment to use.")
@click.argument("directory", type=click.Path(exists=True, file_okay=False), default=os.getcwd())
@requires_project_directory
def edit_values(directory, cname, ename):  # pylint: disable=redefined-outer-name
    base_dir = project.get_base_dir(directory, cname, ename)

    file_path = os.path.join(base_dir, "values.yaml")

    click.edit(filename=file_path)


def main():
    try:
        root()  # pylint: disable=no-value-for-parameter
    except RuntimeError as exc:
        sys.exit(exc)



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os

from click.testing import CliRunner

import boto3
from moto import mock_ssm  # pylint: disable=E0401

from k8t import __license__, __version__
from k8t.cli import root
from k8t.scaffolding import list_available_templates


def test_print_version():
    runner = CliRunner()

    result = runner.invoke(root, ['--version'])
    assert result.exit_code == 0
    assert __version__ in result.output


def test_print_license():
    runner = CliRunner()

    result = runner.invoke(root, ['license'])
    assert result.exit_code == 0
    assert __license__ in result.output


def test_new_project():
    runner = CliRunner()

    with runner.isolated_filesystem():
        result = runner.invoke(root, ['new', 'project', 'test'])
        assert result.exit_code == 0
        assert 'Directory created: test' in result.output
        assert 'File created: test/.k8t' in result.output
        assert 'File created: test/values.yaml' in result.output
        assert 'File created: test/config.yaml' in result.output
        assert os.path.exists('test/.k8t')
        assert os.path.exists('test/values.yaml')
        assert os.path.exists('test/config.yaml')


def test_new_cluster():
    runner = CliRunner()

    with runner.isolated_filesystem():
        open('.k8t', 'w')

        result = runner.invoke(root, ['new', 'cluster', 'cluster-1', '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./clusters/cluster-1' in result.output
        assert 'File created: ./clusters/cluster-1/values.yaml' in result.output
        assert 'File created: ./clusters/cluster-1/config.yaml' in result.output
        assert os.path.exists('./clusters/cluster-1/values.yaml')
        assert os.path.exists('./clusters/cluster-1/config.yaml')


def test_new_environment():
    runner = CliRunner()

    with runner.isolated_filesystem():
        open('.k8t', 'w')

        result = runner.invoke(root, ['new', 'environment', 'staging', '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./environments/staging' in result.output
        assert 'File created: ./environments/staging/values.yaml' in result.output
        assert 'File created: ./environments/staging/config.yaml' in result.output
        assert os.path.exists('./environments/staging/values.yaml')
        assert os.path.exists('./environments/staging/config.yaml')

        os.makedirs('clusters/cluster-1')

        result = runner.invoke(root, ['new', 'environment', '-c', 'cluster-1', 'production', '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./clusters/cluster-1/environments/production' in result.output
        assert 'File created: ./clusters/cluster-1/environments/production/values.yaml' in result.output
        assert 'File created: ./clusters/cluster-1/environments/production/config.yaml' in result.output
        assert os.path.exists('./clusters/cluster-1/environments/production/values.yaml')
        assert os.path.exists('./clusters/cluster-1/environments/production/config.yaml')


def test_new_template():
    template_type = list(list_available_templates())[0]
    runner = CliRunner()

    with runner.isolated_filesystem():
        open('.k8t', 'w')

        result = runner.invoke(root, ['new', 'template', '-n', 'bar', template_type, '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./templates' in result.output
        assert f'Template created: ./templates/bar-{template_type}.yaml.j2' in result.output
        assert os.path.exists(f'./templates/bar-{template_type}.yaml.j2')

        result = runner.invoke(root, ['new', 'template', '-p', 'foo', template_type, '.'])
        assert result.exit_code == 0
        assert f'Template created: ./templates/foo{template_type}.yaml.j2' in result.output
        assert os.path.exists(f'./templates/foo{template_type}.yaml.j2')

        result = runner.invoke(root, ['new', 'template', '-p', 'foo', '-n', 'bar', template_type, '.'])
        assert result.exit_code == 0
        assert f'Template created: ./templates/foobar-{template_type}.yaml.j2' in result.output
        assert os.path.exists(f'./templates/foobar-{template_type}.yaml.j2')

        os.makedirs('environments/staging')

        result = runner.invoke(root, ['new', 'template', '-e', 'staging', template_type, '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./environments/staging/templates' in result.output
        assert f'Template created: ./environments/staging/templates/{template_type}.yaml.j2' in result.output
        assert os.path.exists(f'./environments/staging/templates/{template_type}.yaml.j2')

        os.makedirs('clusters/cluster-1')

        result = runner.invoke(root, ['new', 'template', '-c', 'cluster-1', template_type, '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./clusters/cluster-1/templates' in result.output
        assert f'Template created: ./clusters/cluster-1/templates/{template_type}.yaml.j2' in result.output
        assert os.path.exists(f'./clusters/cluster-1/templates/{template_type}.yaml.j2')

        os.makedirs('clusters/cluster-1/environments/production')

        result = runner.invoke(root, ['new', 'template', '-c', 'cluster-1', '-e', 'production', template_type, '.'])
        assert result.exit_code == 0
        assert 'Directory created: ./clusters/cluster-1/environments/production/templates' in result.output
        assert f'Template created: ./clusters/cluster-1/environments/production/templates/{template_type}.yaml.j2' in result.output
        assert os.path.exists(f'./clusters/cluster-1/environments/production/templates/{template_type}.yaml.j2')

        for template_type in list_available_templates():
            result = runner.invoke(root, ['new', 'template', template_type, '.'])
            assert result.exit_code == 0
            assert f'Template created: ./templates/{template_type}.yaml.j2' in result.output
            assert os.path.exists(f'./templates/{template_type}.yaml.j2')


def test_get_clusters():
    runner = CliRunner()

    result = runner.invoke(root, ['get', 'clusters', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'cluster-1' in result.output
    assert 'cluster-2' in result.output


def test_get_environments():
    runner = CliRunner()

    result = runner.invoke(root, ['get', 'environments', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'some-env' in result.output
    assert 'common-env' in result.output
    assert 'cluster-specific-env' in result.output

    result = runner.invoke(root, ['get', 'environments', '-c', 'cluster-1', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'some-env' not in result.output
    assert 'common-env' in result.output
    assert 'cluster-specific-env' in result.output

    result = runner.invoke(root, ['get', 'environments', '-c', 'cluster-2', 'tests/resources/good'])
    assert result.exit_code == 0
    assert not result.output


def test_get_templates():  # pylint: disable=R0915
    runner = CliRunner()

    result = runner.invoke(root, ['get', 'templates', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' not in result.output
    assert 'cluster-1-template.yaml.j2' not in result.output
    assert 'cluster-1-env-template.yaml.j2' not in result.output
    assert 'composite-template.yaml.j2' not in result.output

    result = runner.invoke(root, ['get', 'templates', '-e', 'common-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' in result.output
    assert 'cluster-1-template.yaml.j2' not in result.output
    assert 'cluster-1-env-template.yaml.j2' not in result.output
    assert 'composite-template.yaml.j2' not in result.output

    result = runner.invoke(root, ['get', 'templates', '-e', 'some-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' not in result.output
    assert 'cluster-1-template.yaml.j2' not in result.output
    assert 'cluster-1-env-template.yaml.j2' not in result.output
    assert 'composite-template.yaml.j2' not in result.output

    result = runner.invoke(root, ['get', 'templates', '-c', 'cluster-1', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' not in result.output
    assert 'cluster-1-template.yaml.j2' in result.output
    assert 'cluster-1-env-template.yaml.j2' not in result.output
    assert 'composite-template.yaml.j2' not in result.output

    result = runner.invoke(root, ['get', 'templates', '-c', 'cluster-1', '-e', 'cluster-specific-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' not in result.output
    assert 'cluster-1-template.yaml.j2' in result.output
    assert 'cluster-1-env-template.yaml.j2' in result.output
    assert 'composite-template.yaml.j2' not in result.output

    result = runner.invoke(root, ['get', 'templates', '-c', 'cluster-1', '-e', 'common-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' in result.output
    assert 'cluster-1-template.yaml.j2' in result.output
    assert 'cluster-1-env-template.yaml.j2' not in result.output
    assert 'composite-template.yaml.j2' not in result.output

    result = runner.invoke(root, ['get', 'templates', '-c', 'cluster-2', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2' in result.output
    assert 'common-env-template.yaml.j2' not in result.output
    assert 'cluster-1-template.yaml.j2' not in result.output
    assert 'cluster-1-env-template.yaml.j2' not in result.output
    assert 'composite-template.yaml.j2' in result.output


def test_validate_successful():  # pylint: disable=R0915
    runner = CliRunner()

    result = runner.invoke(root, ['validate', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' not in result.output
    assert 'composite-template.yaml.j2: ✔' not in result.output

    result = runner.invoke(root, ['validate', '-e', 'common-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' in result.output
    assert 'cluster-1-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' not in result.output
    assert 'composite-template.yaml.j2: ✔' not in result.output

    result = runner.invoke(root, ['validate', '-e', 'some-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' not in result.output
    assert 'composite-template.yaml.j2: ✔' not in result.output

    result = runner.invoke(root, ['validate', '-c', 'cluster-1', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-template.yaml.j2: ✔' in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' not in result.output
    assert 'composite-template.yaml.j2: ✔' not in result.output

    result = runner.invoke(root, ['validate', '-c', 'cluster-1', '-e', 'cluster-specific-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-template.yaml.j2: ✔' in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' in result.output
    assert 'composite-template.yaml.j2: ✔' not in result.output

    result = runner.invoke(root, ['validate', '-c', 'cluster-1', '-e', 'common-env', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' in result.output
    assert 'cluster-1-template.yaml.j2: ✔' in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' not in result.output
    assert 'composite-template.yaml.j2: ✔' not in result.output

    result = runner.invoke(root, ['validate', '-c', 'cluster-2', 'tests/resources/good'])
    assert result.exit_code == 0
    assert 'common-template.yaml.j2: ✔' in result.output
    assert 'common-env-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-template.yaml.j2: ✔' not in result.output
    assert 'cluster-1-env-template.yaml.j2: ✔' not in result.output
    assert 'composite-template.yaml.j2: ✔' in result.output


def test_validate_with_values():
    runner = CliRunner()

    result = runner.invoke(root, ['validate', 'tests/resources/missing_values'])
    assert result.exit_code == 1
    assert 'template.yaml.j2: ✗' in result.output
    assert '- undefined variable: test' in result.output

    result = runner.invoke(root, ['validate', '--value', 'test', 'something', 'tests/resources/missing_values'])
    assert result.exit_code == 0
    assert 'template.yaml.j2: ✔' in result.output

    result = runner.invoke(
        root,
        ['validate', '--value-file', 'tests/resources/missing_values/value-file.yaml', 'tests/resources/missing_values']
    )
    assert result.exit_code == 0
    assert 'template.yaml.j2: ✔' in result.output


def test_validate_failure():
    runner = CliRunner()

    result = runner.invoke(root, ['validate', 'tests/resources/bad'])
    # TODO: Fix and uncomment
    # assert 'filter-template.yaml.j2: ✗' in result.output
    # assert 'nested-value-adding-template.yaml.j2: ✗' in result.output
    # assert 'nested-value-template.yaml.j2: ✗' in result.output
    assert 'secret-template.yaml.j2: ✗' in result.output
    assert 'several-template.yaml.j2: ✗' in result.output
    assert 'value-template.yaml.j2: ✗' in result.output
    assert 'composite-template.yaml.j2: ✗' in result.output
    # assert 'invalid-yaml-template.yaml.j2: ✗' in result.output
    # assert 'composite-invalid-yaml-template.yaml.j2: ✗' in result.output

@mock_ssm
def test_gen():
    client = boto3.client('ssm', region_name='eu-central-1')

    client.put_parameter(
        Name="/cluster-1/test",
        Description="Environment specific simple parameter",
        Value="string_value",
        Type="String",
    )

    runner = CliRunner()

    with open('tests/resources/results/default.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/default-with-cli-values.yaml', 'r') as file:
        result = runner.invoke(
            root, ['gen', '--value', 'foo', '77950e8', 'tests/resources/good']
        )
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/some-env.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', '-e', 'some-env', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/common-env.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', '-e', 'common-env', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/cluster-1.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', '-c', 'cluster-1', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/cluster-1-common-env.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', '-c', 'cluster-1', '-e', 'common-env', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/cluster-1-cluster-specific-env.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', '-c', 'cluster-1', '-e', 'cluster-specific-env', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()

    with open('tests/resources/results/cluster-2.yaml', 'r') as file:
        result = runner.invoke(root, ['gen', '-c', 'cluster-2', 'tests/resources/good'])
        assert result.exit_code == 0
        assert result.output == file.read()


# vim: fenc=utf-8:ts=4:sw=4:expandtab



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
from typing import Set, Tuple

import yaml  # pylint: disable=E0401
from jinja2 import Environment, meta, nodes  # pylint: disable=E0401

from k8t import config

LOGGER = logging.getLogger(__name__)
PROHIBITED_VARIABLE_NAMES = {
    'namespace',
    'cycler',
    'dict',
    'range',
    'lipsum',
    'joiner'
}


class YamlValidationError(Exception):
    """
    Raised if template is rendered into incorrect yaml
    """


def analyze(template_path: str, values: dict, engine: Environment) -> Tuple[Set[str], Set[str], Set[str], bool]:
    template_source = engine.loader.get_source(engine, template_path)[0]
    abstract_syntax_tree = engine.parse(template_source)

    has_secrets = any(getattr(call.node, "name", None) == "get_secret" for call in abstract_syntax_tree.find_all(nodes.Call))
    required_variables = get_variables(abstract_syntax_tree, engine)

    defined_variables = set(values.keys())

    LOGGER.debug(
        "defined variables: %s", defined_variables)
    LOGGER.debug("found required variables in template %s: %s",
                 template_path, required_variables)

    undefined_variables = required_variables.difference(defined_variables)
    unused_variables = defined_variables.difference(required_variables)
    invalid_variables = {
        var for var in required_variables if var in PROHIBITED_VARIABLE_NAMES
    }

    return (undefined_variables - invalid_variables), unused_variables, invalid_variables, has_secrets


def validate(template_path: str, values: dict, engine: Environment) -> bool:
    config_ok = True
    undefined, _, invalid, has_secrets = analyze(template_path, values, engine)

    if undefined:
        LOGGER.error(
            "Undefined variables found: %s", sorted(undefined))

    if invalid:
        LOGGER.error(
            "Invalid variable names found: %s", sorted(invalid))

    if has_secrets:
        if "secrets" not in config.CONFIG:
            LOGGER.error(
                "No configuration for secrets found: %s", config.CONFIG)
            config_ok = False

    return config_ok and not (invalid or undefined)


def get_variables(ast, engine: Environment) -> Set[str]:
    undeclared = meta.find_undeclared_variables(ast)

    defaults = {
        filter.node.name

        for filter in ast.find_all(nodes.Filter)

        if filter.name == "default" and hasattr(filter.node, "name")
    }

    guarded = {
        test.node.name

        for test in ast.find_all(nodes.Test)

        if test.name == 'defined' and hasattr(test.node, "name")
    }

    local = {
        assign.target.name

        for assign in ast.find_all(nodes.Assign)
    }

    return (
        undeclared - (set(engine.globals.keys()) - PROHIBITED_VARIABLE_NAMES) - defaults - guarded - local
    )


def render(template_path: str, values: dict, engine: Environment) -> str:
    output = engine.get_template(template_path).render(values)

    try:
        yaml.safe_load_all(output)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as err:
        raise YamlValidationError(err) from err

    return output



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#
# Copyright © 2020 Clark Germany GmbH
# Author: Aljosha Friemann <aljosha.friemann@clark.de>

import base64
import hashlib
import os
import re
import string
from typing import Any, Optional

from k8t import config, secret_providers

try:
    from secrets import choice
except ImportError:
    from random import SystemRandom

    choice = SystemRandom().choice


def random_password(length: int) -> str:
    return "".join(
        choice(string.ascii_lowercase + string.digits) for _ in range(length)
    )


def envvar(key: str, default=None) -> str:
    return os.environ.get(key, default)


def b64encode(value: Any) -> str:
    result = None

    if isinstance(value, str):
        result = base64.b64encode(value.encode()).decode()
    elif isinstance(value, int):
        result = base64.b64encode(str(value).encode()).decode()
    elif isinstance(value, bytes):
        result = base64.b64encode(value).decode()
    else:
        raise TypeError("invalid input: {}".format(value))

    return result


def b64decode(value: Any) -> str:
    result = None

    if isinstance(value, str):
        result = base64.b64decode(value.encode()).decode()
    elif isinstance(value, bytes):
        result = base64.b64decode(value).decode()
    else:
        raise TypeError("invalid input: {}".format(value))

    return result


def hashf(value, method="sha256"):
    try:
        hash_method = getattr(hashlib, method)()
    except AttributeError as no_hash_method:
        raise RuntimeError("No such hash method: {}".format(method)) from no_hash_method

    if isinstance(value, str):
        hash_method.update(value.encode())
    elif isinstance(value, bytes):
        hash_method.update(value)
    else:
        raise TypeError("invalid input: {}".format(value))

    return hash_method.hexdigest()


def get_secret(key: str, length: int = None) -> str:
    provider_name = config.CONFIG.get("secrets", {}).get("provider")

    if not provider_name:
        raise RuntimeError("Secrets provider not configured.")

    provider_name = str(provider_name).lower()
    try:
        provider = getattr(secret_providers, provider_name)
    except AttributeError as no_secret_provider:
        raise NotImplementedError("secret provider {} does not exist.".format(provider_name)) from no_secret_provider

    return provider(key, length)


def to_bool(value: Any) -> Optional[bool]:
    if value is None or isinstance(value, bool):
        return value

    if isinstance(value, str):
        value = value.lower()

    if value in ('yes', 'on', '1', 'true', 1):
        return True

    return False


def sanitize_label(value: str) -> str:
    """
    source: https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set

    TODO i'm sure there is a smarter way to do this.
    """

    return re.sub(r'(^[^a-z0-9A-Z]|[^a-z0-9A-Z]$|[^a-z0-9A-Z_.-])', 'X', value[:63])



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import boto3
import pytest  # pylint: disable=E0401
from moto import mock_ssm  # pylint: disable=E0401

from k8t import config
from k8t.secret_providers import random, ssm

def test_random():
    config.CONFIG = {"secrets": {"provider": "random"}}
    assert random("/foobar"), random("/foobar")
    assert random("/foobar") != random("/foobaz")

    assert len(random("/foo", 12)) == 12
    with pytest.raises(
        AssertionError, match=r"Secret '/foo' did not have expected length of 3"
    ):
        random("/foo", 3)


@mock_ssm
def test_ssm():
    region = "eu-west-1"
    client = boto3.client("ssm", region_name=region)

    client.put_parameter(
        Name="foo",
        Description="Global parameter",
        Value="global_secret_value",
        Type="SecureString",
        KeyId="alias/aws/ssm",
    )

    client.put_parameter(
        Name="/dev/test1",
        Description="Environment specific simple parameter",
        Value="string_value",
        Type="String",
    )

    client.put_parameter(
        Name="/app/dev/password",
        Description="Environment specific secret.",
        Value="my_secret_value",
        Type="SecureString",
        KeyId="alias/aws/ssm",
    )

    config.CONFIG = {"secrets": {"provider": "ssm", "region": region}}
    assert ssm("foo"), "global_secret_value"
    assert ssm("/dev/test1"), "string_value"
    assert ssm("/app/dev/password"), "my_secret_value"

    with pytest.raises(
        AssertionError,
        match=r"Secret '/app/dev/password' did not have expected length of 3",
    ):
        ssm("/app/dev/password", 3)
    with pytest.raises(
        RuntimeError, match=r"Failed to retrieve secret /Application/non_existent: ..."
    ):
        ssm("/Application/non_existent")
    with pytest.raises(
        RuntimeError, match=r"Failed to retrieve secret /Application/password/: ..."
    ):
        ssm("/Application/password/")

    config.CONFIG = {
        "secrets": {"provider": "ssm", "region": region, "prefix": "/app/dev"}
    }
    assert ssm("/password"), "my_secret_value"

    config.CONFIG = {"secrets": {"provider": "ssm", "region": "eu-central-1"}}
    with pytest.raises(RuntimeError, match=r"Failed to retrieve secret foo: ..."):
        ssm("foo")


@mock_ssm
def test_ssm_default_region():
    region = "eu-central-1"
    client = boto3.client("ssm", region_name=region)

    client.put_parameter(
        Name="foo",
        Description="Global parameter",
        Value="global_secret_value",
        Type="SecureString",
        KeyId="alias/aws/ssm",
    )

    config.CONFIG = {"secrets": {"provider": "ssm"}}
    assert ssm("foo"), "global_secret_value"


# vim: fenc=utf-8:ts=4:sw=4:expandtab



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
from typing import List


def check_directory(path: str) -> bool:
    return os.path.exists(os.path.join(path, ".k8t"))


def get_base_dir(root: str, cluster: str, environment: str) -> str:
    base_path = root

    if cluster is not None:
        base_path = os.path.join(base_path, "clusters", cluster)

        if not os.path.isdir(base_path):
            raise RuntimeError("No such cluster: {}".format(cluster))

    if environment is not None:
        base_path = os.path.join(base_path, "environments", environment)

        if not os.path.isdir(base_path):
            raise RuntimeError("No such environment: {}".format(environment))

    return base_path


# pylint: disable=too-many-arguments
def find_files(root: str, cluster: str, environment: str, name: str, file_ok=True, dir_ok=True) -> List[str]:
    def check(path):
        return (file_ok and os.path.isfile(path)) or (dir_ok and os.path.isdir(root_path))

    files: List[str] = []

    root_path = os.path.join(root, name)

    if check(root_path):
        files.append(root_path)

    env_found = environment is None

    if environment is not None:
        file_path = os.path.join(root, "environments", environment, name)

        if check(file_path):
            files.append(file_path)
            env_found = True

    if cluster is not None:
        cluster_path = os.path.join(root, "clusters", cluster)

        if not os.path.isdir(cluster_path):
            raise RuntimeError("no such cluster: {}".format(cluster))

        file_path = os.path.join(cluster_path, name)

        if check(file_path):
            files.append(file_path)

        if environment is not None:
            file_path = os.path.join(
                cluster_path, "environments", environment, name)

            if check(file_path):
                files.append(file_path)
                env_found = True

    if not env_found:
        raise RuntimeError("no such environment: {}".format(environment))

    return files



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#
# Copyright © 2020 Clark Germany GmbH
# Author: Aljosha Friemann <aljosha.friemann@clark.de>

import random

import pytest  # pylint: disable=E0401
from mock import patch  # pylint: disable=E0401

from k8t import config, secret_providers
from k8t.filters import (b64decode, b64encode, get_secret, hashf,
                         random_password, sanitize_label, to_bool)


def test_b64encode():
    string = "foobar"

    encoded = b64encode(string)

    assert encoded != string
    assert b64decode(encoded) == string


def test_random_password():
    length = int(random.uniform(1, 200))

    assert len(random_password(length)) == length
    assert random_password(length) != random_password(length)


def test_hashf():
    string = "foobar"

    assert hashf(string) != string
    assert hashf(string) == hashf(string)
    assert hashf(string) != hashf("foobaz")


def test_get_secret():
    config.CONFIG = {"secrets": {"provider": "random"}}
    with patch.object(secret_providers, "random") as mock:
        get_secret("any")
        mock.assert_called_with("any", None)
        get_secret("any", 99)
        mock.assert_called_with("any", 99)

    config.CONFIG = {"secrets": {"provider": "ssm"}}
    with patch.object(secret_providers, "ssm") as mock:
        get_secret("any")
        mock.assert_called_with("any", None)
        get_secret("any", 99)
        mock.assert_called_with("any", 99)

    config.CONFIG = {"secrets": {"provider": "nothing"}}
    with pytest.raises(NotImplementedError, match=r"secret provider nothing does not exist."):
        get_secret("any")

    config.CONFIG = {}
    with pytest.raises(RuntimeError, match=r"Secrets provider not configured."):
        get_secret("any")


def test_to_bool():
    assert to_bool(None) is None

    assert to_bool(True) is True
    assert to_bool("Yes") is True
    assert to_bool("yes") is True
    assert to_bool("True") is True
    assert to_bool("true") is True
    assert to_bool("On") is True
    assert to_bool("on") is True
    assert to_bool("1") is True
    assert to_bool(1) is True

    assert to_bool(False) is False
    assert to_bool("False") is False
    assert to_bool("false") is False
    assert to_bool("0") is False
    assert to_bool(0) is False

    # TODO this is a bad one, probably need to raise an error instead
    assert to_bool("anything") is False


def test_sanitize_label():
    # sanity check
    assert sanitize_label("foobar") == "foobar"

    # check replacement for first and last characters
    assert sanitize_label(",foobar-") == "XfoobarX"

    # check valid characters
    assert sanitize_label(",foobar-baz.com") == "Xfoobar-baz.com"

    # check invalid character replacements
    assert sanitize_label(",foobar-baz$.com") == "Xfoobar-bazX.com"

    # check length
    assert len(sanitize_label("x" * 65)) == 63



# -*- coding: utf-8 -*-
# ISC License
#
# Copyright 2019 FL Fintech E GmbH
#
# Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted, provided that the above copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
from typing import Any, Dict, List

from k8t.project import find_files
from k8t.util import deep_merge, load_yaml

LOGGER = logging.getLogger(__name__)


def load_all(root: str, cluster: str, environment: str, method: str) -> Dict[str, Any]:
    values: List[str] = find_files(
        root, cluster, environment, "values.yaml", dir_ok=False)

    LOGGER.debug("using value files: %s", values)

    base = dict()

    if cluster is not None:
        base['cluster'] = cluster

    if environment is not None:
        base['environment'] = environment

    return deep_merge(base, *[load_yaml(f) for f in values], method=method)
